/* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 *  Copyright (c) 2001-2003 Cluster File Systems, Inc.
 *
 *   This file is part of Lustre, http://www.lustre.org.
 *
 *   Lustre is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   Lustre is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Lustre; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * These are the only exported functions, they provide some generic
 * infrastructure for managing object devices
 */

#define DEBUG_SUBSYSTEM S_CLASS
#ifdef __KERNEL__
#include <linux/kmod.h>   /* for request_module() */
#include <linux/module.h>
#include <linux/obd_class.h>
#include <linux/random.h>
#include <linux/slab.h>
#else 
#include <liblustre.h>
#include <linux/obd_class.h>
#include <linux/obd.h>
#endif
#include <linux/lprocfs_status.h>

extern struct list_head obd_types;
kmem_cache_t *obdo_cachep = NULL;
kmem_cache_t *import_cachep = NULL;
kmem_cache_t *export_cachep = NULL;

int (*ptlrpc_put_connection_superhack)(struct ptlrpc_connection *c);
void (*ptlrpc_abort_inflight_superhack)(struct obd_import *imp,
                                        int dying_import);

/*
 * support functions: we could use inter-module communication, but this
 * is more portable to other OS's
 */
static struct obd_type *class_search_type(char *name)
{
        struct list_head *tmp;
        struct obd_type *type;
        CDEBUG(D_INFO, "SEARCH %s\n", name);

        tmp = &obd_types;
        list_for_each(tmp, &obd_types) {
                type = list_entry(tmp, struct obd_type, typ_chain);
                CDEBUG(D_INFO, "TYP %s\n", type->typ_name);
                if (strlen(type->typ_name) == strlen(name) &&
                    strcmp(type->typ_name, name) == 0) {
                        return type;
                }
        }
        return NULL;
}

struct obd_type *class_get_type(char *name)
{
        struct obd_type *type = class_search_type(name);

#ifdef CONFIG_KMOD
        if (!type) {
                if (!request_module(name)) {
                        CDEBUG(D_INFO, "Loaded module '%s'\n", name);
                        type = class_search_type(name);
                } else
                        CDEBUG(D_INFO, "Can't load module '%s'\n", name);
        }
#endif
        if (type)
                __MOD_INC_USE_COUNT(type->typ_ops->o_owner);
        return type;
}

void class_put_type(struct obd_type *type)
{
        LASSERT(type);
        __MOD_DEC_USE_COUNT(type->typ_ops->o_owner);
}

int class_register_type(struct obd_ops *ops, struct lprocfs_vars *vars,
                        char *name)
{
        struct obd_type *type;
        int rc = 0;
        ENTRY;

        LASSERT(strnlen(name, 1024) < 1024);    /* sanity check */

        if (class_search_type(name)) {
                CDEBUG(D_IOCTL, "Type %s already registered\n", name);
                RETURN(-EEXIST);
        }

        rc = -ENOMEM;
        OBD_ALLOC(type, sizeof(*type));
        if (type == NULL)
                RETURN(rc);

        OBD_ALLOC(type->typ_ops, sizeof(*type->typ_ops));
        OBD_ALLOC(type->typ_name, strlen(name) + 1);
        if (type->typ_ops == NULL || type->typ_name == NULL)
                GOTO (failed, rc);

        *(type->typ_ops) = *ops;
        strcpy(type->typ_name, name);
        list_add(&type->typ_chain, &obd_types);

        type->typ_procroot = lprocfs_register(type->typ_name, proc_lustre_root,
                                              vars, type);
        if (IS_ERR(type->typ_procroot)) {
                rc = PTR_ERR(type->typ_procroot);
                type->typ_procroot = NULL;
                list_del(&type->typ_chain);
                GOTO (failed, rc);
        }

        RETURN (0);

 failed:
        if (type->typ_ops != NULL)
                OBD_FREE(type->typ_name, strlen(name) + 1);
        if (type->typ_ops != NULL)
                OBD_FREE (type->typ_ops, sizeof (*type->typ_ops));
        RETURN(rc);
}

int class_unregister_type(char *name)
{
        struct obd_type *type = class_search_type(name);
        ENTRY;

        if (!type) {
                CERROR("unknown obd type\n");
                RETURN(-EINVAL);
        }

        if (type->typ_refcnt) {
                CERROR("type %s has refcount (%d)\n", name, type->typ_refcnt);
                /* This is a bad situation, let's make the best of it */
                /* Remove ops, but leave the name for debugging */
                OBD_FREE(type->typ_ops, sizeof(*type->typ_ops));
                RETURN(-EBUSY);
        }

        if (type->typ_procroot) {
                lprocfs_remove(type->typ_procroot);
                type->typ_procroot = NULL;
        }

        list_del(&type->typ_chain);
        OBD_FREE(type->typ_name, strlen(name) + 1);
        if (type->typ_ops != NULL)
                OBD_FREE(type->typ_ops, sizeof(*type->typ_ops));
        OBD_FREE(type, sizeof(*type));
        RETURN(0);
} /* class_unregister_type */

int class_name2dev(char *name)
{
        int res = -1;
        int i;

        if (!name)
                return -1;

        for (i = 0; i < MAX_OBD_DEVICES; i++) {
                struct obd_device *obd = &obd_dev[i];
                if (obd->obd_name && strcmp(name, obd->obd_name) == 0) {
                        res = i;
                        return res;
                }
        }

        return res;
}

int class_uuid2dev(struct obd_uuid *uuid)
{
        int res = -1;
        int i;

        for (i = 0; i < MAX_OBD_DEVICES; i++) {
                struct obd_device *obd = &obd_dev[i];
                if (strncmp(uuid->uuid, obd->obd_uuid.uuid, sizeof(obd->obd_uuid.uuid)) == 0) {
                        res = i;
                        return res;
                }
        }

        return res;
}


struct obd_device *class_uuid2obd(struct obd_uuid *uuid)
{
        int i;

        for (i = 0; i < MAX_OBD_DEVICES; i++) {
                struct obd_device *obd = &obd_dev[i];
                if (strncmp(uuid->uuid, obd->obd_uuid.uuid, sizeof(obd->obd_uuid.uuid)) == 0)
                        return obd;
        }

        return NULL;
}

void obd_cleanup_caches(void)
{
        int rc;
        ENTRY;
        if (obdo_cachep) {
                rc = kmem_cache_destroy(obdo_cachep);
                if (rc)
                        CERROR("Cannot destory ll_obdo_cache\n");
                obdo_cachep = NULL;
        }
        if (import_cachep) {
                rc = kmem_cache_destroy(import_cachep);
                if (rc)
                        CERROR("Cannot destory ll_import_cache\n");
                import_cachep = NULL;
        }
        if (export_cachep) {
                rc = kmem_cache_destroy(export_cachep);
                if (rc)
                        CERROR("Cannot destory ll_export_cache\n");
                export_cachep = NULL;
        }
        EXIT;
}

int obd_init_caches(void)
{
        ENTRY;
        LASSERT(obdo_cachep == NULL);
        obdo_cachep = kmem_cache_create("ll_obdo_cache", sizeof(struct obdo),
                                        0, 0, NULL, NULL);
        if (!obdo_cachep)
                GOTO(out, -ENOMEM);

        LASSERT(export_cachep == NULL);
        export_cachep = kmem_cache_create("ll_export_cache",
                                          sizeof(struct obd_export),
                                          0, 0, NULL, NULL);
        if (!export_cachep)
                GOTO(out, -ENOMEM);

        LASSERT(import_cachep == NULL);
        import_cachep = kmem_cache_create("ll_import_cache",
                                          sizeof(struct obd_import),
                                          0, 0, NULL, NULL);
        if (!import_cachep)
                GOTO(out, -ENOMEM);

        RETURN(0);
 out:
        obd_cleanup_caches();
        RETURN(-ENOMEM);

}

/* map connection to client */
struct obd_export *class_conn2export(struct lustre_handle *conn)
{
        struct obd_export *export;
        ENTRY;

        if (!conn) {
                CDEBUG(D_CACHE, "looking for null handle\n");
                RETURN(NULL);
        }

        if (conn->addr == -1) {  /* this means assign a new connection */
                CDEBUG(D_CACHE, "want a new connection\n");
                RETURN(NULL);
        }

        if (!conn->addr) {
                CDEBUG(D_CACHE, "looking for null addr\n");
                fixme();
                RETURN(NULL);
        }

        CDEBUG(D_IOCTL, "looking for export addr "LPX64" cookie "LPX64"\n",
               conn->addr, conn->cookie);
        export = (struct obd_export *) (unsigned long)conn->addr;
        if (!kmem_cache_validate(export_cachep, (void *)export))
                RETURN(NULL);

        if (export->exp_cookie != conn->cookie)
                RETURN(NULL);
        RETURN(export);
} /* class_conn2export */

struct obd_device *class_conn2obd(struct lustre_handle *conn)
{
        struct obd_export *export;
        export = class_conn2export(conn);
        if (export)
                return export->exp_obd;
        fixme();
        return NULL;
}

struct obd_import *class_conn2cliimp(struct lustre_handle *conn)
{
        return &class_conn2obd(conn)->u.cli.cl_import;
}

struct obd_import *class_conn2ldlmimp(struct lustre_handle *conn)
{
        return &class_conn2export(conn)->exp_ldlm_data.led_import;
}

struct obd_export *class_new_export(struct obd_device *obddev)
{
        struct obd_export *export;

        PORTAL_SLAB_ALLOC(export, export_cachep, sizeof(*export));
        if (!export) {
                CERROR("no memory! (minor %d)\n", obddev->obd_minor);
                return NULL;
        }

        get_random_bytes(&export->exp_cookie, sizeof(export->exp_cookie));
        export->exp_obd = obddev;
        /* XXX this should be in LDLM init */
        INIT_LIST_HEAD(&export->exp_ldlm_data.led_held_locks);
        INIT_LIST_HEAD(&export->exp_conn_chain);
        spin_lock(&obddev->obd_dev_lock);
        list_add(&export->exp_obd_chain, &export->exp_obd->obd_exports);
        spin_unlock(&obddev->obd_dev_lock);
        return export;
}

void class_destroy_export(struct obd_export *exp)
{
        LASSERT(exp->exp_cookie != DEAD_HANDLE_MAGIC);

        CDEBUG(D_IOCTL, "destroying export %p/%s\n", exp,
               exp->exp_client_uuid.uuid);

        spin_lock(&exp->exp_obd->obd_dev_lock);
        list_del(&exp->exp_obd_chain);
        spin_unlock(&exp->exp_obd->obd_dev_lock);

        /* XXXshaver no connection here... */
        if (exp->exp_connection)
                spin_lock(&exp->exp_connection->c_lock);
        list_del(&exp->exp_conn_chain);
        if (exp->exp_connection) {
                spin_unlock(&exp->exp_connection->c_lock);
                ptlrpc_put_connection_superhack(exp->exp_connection);
        }

        /* Abort any inflight DLM requests and NULL out their (about to be
         * freed) import. */
        if (exp->exp_ldlm_data.led_import.imp_obd)
                ptlrpc_abort_inflight_superhack(&exp->exp_ldlm_data.led_import,
                                                1);

        PORTAL_SLAB_FREE(exp, export_cachep, sizeof(*exp));
}

/* a connection defines an export context in which preallocation can
   be managed. */
int class_connect(struct lustre_handle *exporth, struct obd_device *obd,
                  struct obd_uuid *cluuid)
{
        struct obd_export * export;
        if (exporth == NULL) {
                LBUG();
                return -EINVAL;
        }

        if (obd == NULL) {
                LBUG();
                return -EINVAL;
        }

        if (cluuid == NULL) {
                LBUG();
                return -EINVAL;
        }

        export = class_new_export(obd);
        if (!export)
                return -ENOMEM;

        exporth->addr = (__u64) (unsigned long)export;
        exporth->cookie = export->exp_cookie;
        memcpy(&export->exp_client_uuid, cluuid, sizeof(export->exp_client_uuid));

        CDEBUG(D_IOCTL, "connect: addr %Lx cookie %Lx\n",
               (long long)exporth->addr, (long long)exporth->cookie);
        return 0;
}

int class_disconnect(struct lustre_handle *conn)
{
        struct obd_export *export;
        ENTRY;

        if (!(export = class_conn2export(conn))) {
                fixme();
                CDEBUG(D_IOCTL, "disconnect: attempting to free "
                       "nonexistent client "LPX64"\n", conn->addr);
                RETURN(-EINVAL);
        }

        CDEBUG(D_IOCTL, "disconnect: addr %Lx cookie %Lx\n",
                       (long long)conn->addr, (long long)conn->cookie);

        class_destroy_export(export);

        RETURN(0);
}

void class_disconnect_all(struct obd_device *obddev)
{
        int again = 1;

        while (again) {
                spin_lock(&obddev->obd_dev_lock);
                if (!list_empty(&obddev->obd_exports)) {
                        struct obd_export *export;
                        struct lustre_handle conn;
                        int rc;

                        export = list_entry(obddev->obd_exports.next,
                                            struct obd_export,
                                            exp_obd_chain);
                        conn.addr = (__u64)(unsigned long)export;
                        conn.cookie = export->exp_cookie;
                        spin_unlock(&obddev->obd_dev_lock);
                        CERROR("force disconnecting %s:%s export %p\n",
                               export->exp_obd->obd_type->typ_name,
                               export->exp_connection ?
                               (char *)export->exp_connection->c_remote_uuid.uuid :
                               "<unconnected>", export);
                        rc = obd_disconnect(&conn);
                        if (rc < 0) {
                                /* AED: not so sure about this...  We can't
                                 * loop here forever, yet we shouldn't leak
                                 * exports on a struct we will soon destroy.
                                 */
                                CERROR("destroy export %p with err: rc = %d\n",
                                       export, rc);
                                class_destroy_export(export);
                        }
                } else {
                        spin_unlock(&obddev->obd_dev_lock);
                        again = 0;
                }
        }
}

#if 0

/* FIXME: Data is a space- or comma-separated list of device IDs.  This will
 * have to change. */
int class_multi_setup(struct obd_device *obddev, uint32_t len, void *data)
{
        int count, rc;
        char *p;
        ENTRY;

        for (p = data, count = 0; p < (char *)data + len; count++) {
                char *end;
                int tmp = simple_strtoul(p, &end, 0);

                if (p == end) {
                        CERROR("invalid device ID starting at: %s\n", p);
                        GOTO(err_disconnect, rc = -EINVAL);
                }

                if (tmp < 0 || tmp >= MAX_OBD_DEVICES) {
                        CERROR("Trying to sub dev %d  - dev no too large\n",
                               tmp);
                        GOTO(err_disconnect, rc  = -EINVAL);
                }

                rc = obd_connect(&obddev->obd_multi_conn[count], &obd_dev[tmp]);
                if (rc) {
                        CERROR("cannot connect to device %d: rc = %d\n", tmp,
                               rc);
                        GOTO(err_disconnect, rc);
                }

                CDEBUG(D_INFO, "target OBD %d is of type %s\n", count,
                       obd_dev[tmp].obd_type->typ_name);

                p = end + 1;
        }

        obddev->obd_multi_count = count;

        RETURN(0);

 err_disconnect:
        for (count--; count >= 0; count--)
                obd_disconnect(&obddev->obd_multi_conn[count]);
        return rc;
}

/*
 *    remove all connections to this device
 *    close all connections to lower devices
 *    needed for forced unloads of OBD client drivers
 */
int class_multi_cleanup(struct obd_device *obddev)
{
        int i;

        for (i = 0; i < obddev->obd_multi_count; i++) {
                int rc;
                struct obd_device *obd =
                        class_conn2obd(&obddev->obd_multi_conn[i]);

                if (!obd) {
                        CERROR("no such device [i %d]\n", i);
                        RETURN(-EINVAL);
                }

                rc = obd_disconnect(&obddev->obd_multi_conn[i]);
                if (rc)
                        CERROR("disconnect failure %d\n", obd->obd_minor);
        }
        return 0;
}
#endif
